import os
from copy import deepcopy
from typing import Dict, Any, List, Tuple

import mcp

from fibery_mcp_server.fibery_client import FiberyClient, Schema, Database

query_tool_name = "query_database"


def query_tool() -> mcp.types.Tool:
    with open(os.path.join(os.path.dirname(os.path.abspath(__file__)), "descriptions", "query"), "r") as file:
        description = file.read()

    return mcp.types.Tool(
        name=query_tool_name,
        description=description,
        inputSchema={
            "type": "object",
            "properties": {
                "q_from": {
                    "type": "string",
                    "description": 'Specifies the entity type in "Space/Type" format (e.g., "Product Management/feature", "Product Management/Insight")',
                },
                "q_select": {
                    "type": "object",
                    "description": "\n".join(
                        [
                            "Defines what fields to retrieve. Can include:",
                            '  - Primitive fields using format {"AliasName": "FieldName"} (i.e. {"Name": "Product Management/Name"})',
                            '  - Related entity fields using format {"AliasName": ["Related entity", "related entity field"]} (i.e. {"Secret": ["Product Management/Description", "Collaboration~Documents/secret"]}). Careful, does not work with 1-* connection!',
                            'To work with 1-* relationships, you can use sub-querying: {"AliasName": {"q/from": "Related type", "q/select": {"AliasName 2": "fibery/id"}, "q/limit": 50}}',
                            "AliasName can be of any arbitrary value.",
                        ]
                    ),
                },
                "q_where": {
                    "type": "object",
                    "description": "\n".join(
                        [
                            'Filter conditions in format [operator, [field_path], value] or ["q/and"|"q/or", ...conditions]. Common usages:',
                            '- Simple comparison: ["=", ["field", "path"], "$param"]. You cannot pass value of $param directly in where clause. Use params object instead. Pay really close attention to it as it is not common practice, but that\'s how it works in our case!',
                            '- Logical combinations: ["q/and", ["<", ["field1"], "$param1"], ["=", ["field2"], "$param2"]]',
                            "- Available operators: =, !=, <, <=, >, >=, q/contains, q/not-contains, q/in, q/not-in",
                        ]
                    ),
                },
                "q_order_by": {
                    "type": "object",
                    "description": 'List of sorting criteria in format {"field1": "q/asc", "field2": "q/desc"}',
                },
                "q_limit": {
                    "type": "integer",
                    "description": "Number of results per page (defaults to 50). Maximum allowed value is 1000",
                },
                "q_offset": {
                    "type": "integer",
                    "description": "Number of results to skip. Mainly used in combination with limit and orderBy for pagination.",
                },
                "q_params": {
                    "type": "object",
                    "description": 'Dictionary of parameter values referenced in where using "$param" syntax. For example, {$fromDate: "2025-01-01"}',
                },
            },
            "required": ["q_from", "q_select"],
        },
    )


def parse_q_order_by(q_order_by: Dict[str, str] | None) -> List[Tuple[List[str], str]] | None:
    if not q_order_by:
        return None
    return [([field], q_order) for field, q_order in q_order_by.items()]


def get_rich_text_fields(q_select: Dict[str, Any], database: Database) -> Tuple[List[Dict[str, Any]], Dict[str, Any]]:
    rich_text_fields = []
    safe_q_select = deepcopy(q_select)
    for field_alias, field_name in safe_q_select.items():
        if not isinstance(field_name, str):
            if isinstance(field_name, list):
                field_name = field_name[0]
        if database.fields_by_name().get(field_name, None).is_rich_text():
            rich_text_fields.append({"alias": field_alias, "name": field_name})
            safe_q_select[field_alias] = [field_name, "Collaboration~Documents/secret"]
    return rich_text_fields, safe_q_select


async def handle_query(fibery_client: FiberyClient, arguments: Dict[str, Any]) -> List[mcp.types.TextContent]:
    q_from, q_select = arguments["q_from"], arguments["q_select"]

    schema: Schema = await fibery_client.get_schema()
    database = schema.databases_by_name()[arguments["q_from"]]
    rich_text_fields, safe_q_select = get_rich_text_fields(q_select, database)

    base = {
        "q/from": q_from,
        "q/select": safe_q_select,
        "q/limit": arguments.get("q_limit", 50),
    }
    optional = {
        k: v
        for k, v in {
            "q/where": arguments.get("q_where", None),
            "q/order-by": parse_q_order_by(arguments.get("q_order_by", None)),
            "q/offset": arguments.get("q_offset", None),
        }.items()
        if v is not None
    }
    query = base | optional

    commandResult = await fibery_client.query(query, arguments.get("q_params", None))

    if not commandResult.success:
        return [mcp.types.TextContent(type="text", text=str(commandResult))]

    for i, entity in enumerate(commandResult.result):
        for field in rich_text_fields:
            secret = entity.get(field["alias"], None)
            if not secret:
                return [
                    mcp.types.TextContent(
                        type="text", text=f"Unable to get document content for entity {entity}. Field: {field}"
                    )
                ]
            entity[field["alias"]] = await fibery_client.get_document_content(secret)
    return [mcp.types.TextContent(type="text", text=str(commandResult))]



import os
from copy import deepcopy
from typing import List, Dict, Any, Tuple

import mcp

from fibery_mcp_server.fibery_client import FiberyClient, Schema, Database
from fibery_mcp_server.utils import str_to_bool

update_entity_tool_name = "update_entity"


def update_entity_tool() -> mcp.types.Tool:
    with open(os.path.join(os.path.dirname(os.path.abspath(__file__)), "descriptions", "update_entity"), "r") as file:
        description = file.read()

    return mcp.types.Tool(
        name=update_entity_tool_name,
        description=description,
        inputSchema={
            "type": "object",
            "properties": {
                "database": {
                    "type": "string",
                    "description": "Fibery Database where to update an entity.",
                },
                "entity": {
                    "type": "object",
                    "description": "\n".join(
                        [
                            'Dictionary that defines what fields to set in format {"FieldName": value} (i.e. {"Product Management/Name": "My new entity"}).',
                            'Exception are document fields. For them you must specify append (boolean, whether to append to current content) and content itself: {"Product Management/Description": {"append": true, "content": "Additional info"}}',
                        ]
                    ),
                },
            },
            "required": ["database", "entity"],
        },
    )


async def process_fields(
    fibery_client: FiberyClient, schema: Schema, database: Database, fields: Dict[str, Any]
) -> Tuple[List[Dict[str, Any]], Dict[str, Any]]:
    rich_text_fields = []
    safe_fields = deepcopy(fields)
    for field_name, field_value in fields.items():
        # process rich-text fields
        if database.fields_by_name().get(field_name, None).is_rich_text():
            rich_text_fields.append(
                {"name": field_name, "append": str_to_bool(field_value["append"]), "value": field_value["content"]}
            )
            safe_fields.pop(field_name)

        # process enum fields
        field_type = database.fields_by_name().get(field_name, None).type
        if schema.databases_by_name()[field_type].is_enum():
            enum_values_response = await fibery_client.get_enum_values(field_type)
            enum_values = enum_values_response.result
            safe_fields[field_name] = {"fibery/id": next(filter(lambda e: e["Name"] == field_value, enum_values))["Id"]}

    return rich_text_fields, safe_fields


async def handle_update_entity(fibery_client: FiberyClient, arguments: Dict[str, Any]) -> List[mcp.types.TextContent]:
    database_name: str = arguments.get("database")
    entity: Dict[str, Any] = arguments.get("entity")

    if not database_name:
        return [mcp.types.TextContent(type="text", text="Error: database is not provided.")]

    if not entity:
        return [mcp.types.TextContent(type="text", text="Error: entity is not provided.")]

    if not entity["fibery/id"]:
        return [
            mcp.types.TextContent(
                type="text", text="Error: entity id is not provided. Use 'fibery/id' field to set it."
            )
        ]

    schema = await fibery_client.get_schema()
    database = schema.databases_by_name()[database_name]
    if not database:
        return [mcp.types.TextContent(type="text", text=f"Error: database {database_name} was not found.")]
    rich_text_fields, safe_entity = await process_fields(fibery_client, schema, database, entity)

    update_result = await fibery_client.update_entity(database_name, safe_entity)

    if not update_result.success:
        return [mcp.types.TextContent(type="text", text=str(update_result))]

    if len(rich_text_fields) > 0:
        secrets_response = await fibery_client.query(
            {
                "q/from": database_name,
                "q/select": {
                    field["name"]: [field["name"], "Collaboration~Documents/secret"] for field in rich_text_fields
                },
                "q/limit": 1,
                "q/where": ["=", ["fibery/id"], "$id"],
            },
            {"$id": safe_entity["fibery/id"]},
        )

        for field, secret_response in zip(rich_text_fields, secrets_response.result):
            secret = secret_response.get(field["name"], None)
            if not secret:
                return [
                    mcp.types.TextContent(
                        type="text", text=f"Error: entity created, but could you populate document {field['name']}"
                    )
                ]
            doc_result = await fibery_client.create_or_update_document(secret, field["value"], append=field["append"])
            if not doc_result.success:
                return [mcp.types.TextContent(type="text", text=str(doc_result))]

    public_id = await fibery_client.get_public_id_by_id(database_name, safe_entity["fibery/id"])
    url = fibery_client.compose_url(database_name.split("/")[0], database_name.split("/")[1], public_id)
    return [mcp.types.TextContent(type="text", text=str(f"Entity updated successfully. URL: {url}"))]



import os
import sys
import pytest
from uuid import uuid4
from dotenv import load_dotenv

load_dotenv()
sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), "..")))

from src.fibery_mcp_server.fibery_client import FiberyClient

pytestmark = pytest.mark.skipif(
    not os.environ.get("FIBERY_HOST") or not os.environ.get("FIBERY_API_TOKEN"),
    reason="FIBERY_HOST or FIBERY_API_TOKEN environment variables not set",
)
__fibery_host, __fibery_api_token = (
    os.environ.get("FIBERY_HOST"),
    os.environ.get("FIBERY_API_TOKEN"),
)


async def test_get_schema() -> None:
    """Test the get_schema function"""
    fibery_client = FiberyClient(__fibery_host, __fibery_api_token)
    schema = await fibery_client.get_schema()

    assert schema is not None, "No schema returned"
    assert len(schema.databases) > 0, "Schema does not contain 'types' field"


async def test_create_entity() -> None:
    """Test the create_entity function"""
    fibery_client = FiberyClient(__fibery_host, __fibery_api_token)
    fibery_id = str(uuid4())
    creation_result = await fibery_client.create_entity(
        "Product Management/Item", {"fibery/id": fibery_id, "Product Management/Name": "Test"}
    )

    assert creation_result.success is True, "Entity creation failed"

    deletion_result = await fibery_client.delete_entity("Product Management/Item", fibery_id)
    assert deletion_result.success is True, "Entity deletion failed"



from typing import Dict, Any, List
from dataclasses import dataclass

import httpx


class Field:
    def __init__(self, raw_field: Dict[str, Any]):
        self.__raw_field = raw_field
        self.__raw_meta = raw_field.get("fibery/meta", {})

    def is_primitive(self) -> bool:
        return self.__raw_meta.get("fibery/primitive?", False)

    def is_collection(self) -> bool:
        return self.__raw_meta.get("fibery/collection?", False)

    def is_title(self) -> bool:
        return self.__raw_meta.get("ui/title?", False)

    def is_hidden(self) -> bool:
        return self.__raw_meta.get("ui/hidden?", False)

    def is_rich_text(self) -> bool:
        return self.__raw_field.get("fibery/type", None) == "Collaboration~Documents/Document"

    def is_workflow(self) -> bool:
        return self.__raw_field.get("fibery/name", None) == "workflow/state"

    @property
    def type(self) -> str:
        return self.__raw_field["fibery/type"]

    @property
    def primitive_type(self) -> str:
        return self.__raw_field["fibery/type"].split("/")[-1]

    @property
    def name(self) -> str:
        return self.__raw_field["fibery/name"]

    @property
    def title(self) -> str:
        return self.__raw_field["fibery/name"].split("/")[-1].title()


class Database:
    def __init__(self, raw_database: Dict[str, Any]):
        self.__raw_database = raw_database
        self.__raw_meta = raw_database.get("fibery/meta", {})
        self.__fields: List[Field] = [Field(raw_field) for raw_field in raw_database["fibery/fields"]]

    def is_primitive(self) -> bool:
        return self.__raw_meta.get("fibery/primitive?", False)

    def is_enum(self) -> bool:
        return self.__raw_meta.get("fibery/enum?", False)

    def include_database(self) -> bool:
        return not (
            self.name.startswith("fibery/")
            or self.name.startswith("Collaboration~Documents")
            or self.name.endswith("-mixin")
            or self.name == "workflow/workflow"
        )

    def fields_by_name(self) -> Dict[str, Field]:
        return {field.name: field for field in self.__fields}

    @property
    def name(self) -> str:
        return self.__raw_database["fibery/name"]

    @property
    def fields(self) -> List[Field]:
        return self.__fields


class Schema:
    def __init__(self, raw_schema: Dict[str, Any]):
        self.__raw_schema: Dict[str, Any] = raw_schema
        self.__databases: List[Database] = [Database(raw_db) for raw_db in raw_schema["fibery/types"]]

    def databases_by_name(self) -> Dict[str, Database]:
        return {db.name: db for db in self.__databases}

    def include_databases_from_schema(self) -> List[Database]:
        if not self.__databases:
            return []

        databases: List[Database] = []

        for database in filter(lambda db: db.include_database(), self.__databases):
            databases.append(database)
        return databases

    @property
    def databases(self) -> List[Database]:
        return self.__databases


@dataclass
class CommandResponse:
    success: bool
    result: List[Dict[str, Any]] | Dict[str, Any]


@dataclass
class GetDocumentResponse:
    secret: str
    content: str


@dataclass
class CreateDocumentResponse:
    success: bool
    message: str


def normalize_str(s: str) -> str:
    return s.replace(" ", "_").replace("-", "_")


class FiberyClient:
    def __init__(self, fibery_host: str, fibery_api_token: str, fibery_https: bool = True):
        if not fibery_host:
            raise ValueError("Fibery host not provided. Set FIBERY_HOST environment variable.")

        if not fibery_api_token:
            raise ValueError("Fibery API token not provided. Set FIBERY_API_TOKEN environment variable.")

        self.__fibery_host: str = fibery_host
        self.__fibery_api_token: str = fibery_api_token
        self.__fibery_https: bool = fibery_https

    async def fetch_from_fibery(
        self,
        url: str,
        method: str = "GET",
        json_data: Any = None,
        params: Dict[str, str] = None,
    ) -> Dict[str, Any]:
        """
        Generic function to fetch data from Fibery API

        Args:
            url: API endpoint path
            method: HTTP method
            params: Query parameters
            json_data: JSON body of the request

        Returns:
            Response data and metadata
        """

        base_url = f"https://{self.__fibery_host}"
        headers = {
            "Content-Type": "application/json",
            "Authorization": f"Bearer {self.__fibery_api_token}",
        }

        async with httpx.AsyncClient(base_url=base_url, headers=headers, timeout=30.0) as client:
            if method == "GET":
                response = await client.get(url, params=params)
            elif method == "POST":
                response = await client.post(url, json=json_data, params=params)
            else:
                raise ValueError(f"Unsupported HTTP method: {method}")

            response.raise_for_status()

            return {
                "data": response.json() if response.content else None,
            }

    async def get_schema(self) -> Schema:
        """
        Returns:
            Processed Fibery schema
        """
        result = await self.fetch_from_fibery(
            "/api/schema",
            method="GET",
            params={"with-description": "true", "with-soft-deleted": "false"},
        )

        schema_data = result["data"]
        return Schema(schema_data)

    async def execute_command(self, command: str, args: Dict[str, Any]) -> CommandResponse:
        result = await self.fetch_from_fibery(
            "/api/commands",
            method="POST",
            json_data=[
                {
                    "command": command,
                    "args": args,
                },
            ],
        )

        result = result["data"][0]
        return CommandResponse(result["success"], result["result"])

    async def query(self, query: Dict[str, Any], params: Dict[str, Any] | None) -> CommandResponse:
        return await self.execute_command("fibery.entity/query", {"query": query, "params": params})

    async def get_enum_values(self, database_name: str) -> CommandResponse:
        result = await self.fetch_from_fibery(
            "/api/commands",
            method="POST",
            json_data=[
                {
                    "command": "fibery.entity/query",
                    "args": {
                        "query": {
                            "q/from": database_name,
                            "q/select": {"Id": ["fibery/id"], "Name": ["enum/name"]},
                            "q/limit": 100,
                        },
                        "params": {},
                    },
                },
            ],
        )

        result = result["data"][0]
        return CommandResponse(result["success"], result["result"])

    async def get_document_content(self, secret: str) -> str:
        result = await self.fetch_from_fibery(
            f"api/documents/{secret}?format=md",
            method="GET",
        )
        result = result["data"]
        return GetDocumentResponse(result["secret"], result["content"]).content

    async def create_or_update_document(
        self, secret: str, content: str, append: bool = False
    ) -> CreateDocumentResponse:
        result = await self.fetch_from_fibery(
            "/api/documents/commands",
            "POST",
            {
                "command": "create-or-update-documents" if not append else "create-or-append-documents",
                "args": [{"secret": secret, "content": content}],
            },
        )
        result_parsed: bool | Dict[str, Any] = result["data"]
        if result_parsed is True:
            return CreateDocumentResponse(True, "Document created/updated successfully")
        return CreateDocumentResponse(False, result_parsed.get("message", "Failed to create/update document."))

    async def create_entity(self, database: str, entity: Dict[str, Any]) -> CommandResponse:
        return await self.execute_command(
            "fibery.entity/create",
            {
                "type": database,
                "entity": entity,
            },
        )

    async def create_entities_batch(self, database: str, entities: List[Dict[str, Any]]) -> CommandResponse:
        return await self.execute_command(
            "fibery.command/batch",
            {
                "commands": list(map(lambda entity: {
                    "command": "fibery.entity/create",
                    "args": {
                        "type": database,
                        "entity": entity
                    }
                }, entities)),
            },
        )

    async def update_entity(self, database: str, entity: Dict[str, Any]) -> CommandResponse:
        return await self.execute_command(
            "fibery.entity/update",
            {
                "type": database,
                "entity": entity,
            },
        )

    async def delete_entity(self, database: str, fibery_id: str) -> CommandResponse:
        return await self.execute_command(
            "fibery.entity/delete",
            {
                "type": database,
                "entity": {
                    "fibery/id": fibery_id,
                },
            },
        )

    async def get_public_id_by_id(self, database: str, fibery_id: str) -> str | None:
        result = await self.query(
            {
                "q/from": database,
                "q/select": {"Public Id": "fibery/public-id"},
                "q/where": ["=", ["fibery/id"], "$id"],
                "q/limit": 1,
            },
            {"$id": fibery_id},
        )
        if not result.success:
            return None
        return str(result.result[0]["Public Id"])

    def compose_url(self, space: str, database: str, public_id: str) -> str:
        return f"{'https' if self.__fibery_https else 'http'}://{self.__fibery_host}/{normalize_str(space)}/{normalize_str(database)}/{public_id}"



import os
import sys
import pytest
from uuid import uuid4
from dotenv import load_dotenv

load_dotenv()
sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), "..")))

from src.fibery_mcp_server.fibery_client import FiberyClient
from src.fibery_mcp_server.tools.create_entity import handle_create_entity
from src.fibery_mcp_server.tools.update_entity import handle_update_entity

pytestmark = pytest.mark.skipif(
    not os.environ.get("FIBERY_HOST") or not os.environ.get("FIBERY_API_TOKEN"),
    reason="FIBERY_HOST or FIBERY_API_TOKEN environment variables not set",
)
__fibery_host, __fibery_api_token = (
    os.environ.get("FIBERY_HOST"),
    os.environ.get("FIBERY_API_TOKEN"),
)


async def test_create_entity() -> None:
    """Test the create_entity function"""
    fibery_client = FiberyClient(__fibery_host, __fibery_api_token)
    creation_result = await handle_create_entity(
        fibery_client,
        {
            "database": "Product Management/Item",
            "entity": {
                "workflow/state": "To Do",
                "Product Management/Name": "Image Fourier Transform Understanding",
                "Product Management/Description": """A comprehensive summary of 2D Fourier Transform concepts for image processing, including:

- Basic principles of Fourier Transform
- Comparison between 1D (audio) and 2D (image) transforms
- Explanation of frequency domain representation
- Understanding of k and l frequency indices
- Visualization techniques for FFT output
- Edge effects and window functions
- Alignment of frequency patterns with image features
- Practical applications in image processing

This document includes explanations suitable for different audiences from beginners to technical specialists.""",
            },
        },
    )

    print(creation_result)


async def test_update_entity() -> None:
    """Test the create_entity function"""
    fibery_client = FiberyClient(__fibery_host, __fibery_api_token)
    update_result = await handle_update_entity(
        fibery_client,
        {
            "database": "Product Management/Item",
            "entity": {
                "fibery/id": "d3e27780-0ea9-11ef-ab49-df048edba921",
                "workflow/state": "Done",
                "Product Management/Name": "UPDATED 2!",
                "Product Management/Description": {"append": True, "content": " - Updated description"},
            },
        },
    )

    print(update_result)



from copy import deepcopy
from dataclasses import dataclass
from typing import List, Tuple, Dict, Any

from .fibery_client import FiberyClient, Schema, Database, Field


@dataclass
class PrettyField:
    title: str
    name: str
    type: str


def get_ref(schema: Schema, field: Field) -> Database | None:
    if field.is_primitive():
        return None

    ref_database = schema.databases_by_name().get(field.type, None)
    if not ref_database or ref_database.is_primitive():
        return None

    return ref_database


def map_enum_values(enum_values: List[Dict[str, Any]]) -> str:
    return ", ".join([f'"{value["Name"]}"' for value in enum_values])


async def prettify_fields(
    fibery_client: FiberyClient, schema: Schema, database: Database, collect_external_databases: bool = False
) -> Tuple[List[PrettyField], List[Database]]:
    fields = database.fields

    pretty_fields = []
    external_databases: List[Database] = []

    for field in fields:
        if field.is_hidden():
            continue
        title = field.title
        name = field.name
        field_type = field.type

        ref_database = get_ref(schema, field)
        type_str = field.primitive_type if field.is_primitive() else field_type
        if field_type == "fibery/rank":
            type_str = "int"
        if field.is_rich_text():
            type_str = "fibery/document"
        if database.is_enum() and field.is_title():
            enum_values_response = await fibery_client.get_enum_values(database.name)
            type_str += f" # available values: {map_enum_values(enum_values_response.result)}"
        if database.name.split("/")[0] == "workflow" and field.title == "Type":
            type_str += ' # available values: "Not started", "Started", "Finished"'
        if ref_database and not field.is_rich_text():
            type_str = field_type if not field.is_collection() else f"Collection({field_type})"
            if (
                collect_external_databases
                and ref_database.name != database.name
                and ref_database.name not in map(lambda db: db.name, external_databases)
            ):
                external_databases.append(ref_database)
        pretty_fields.append(PrettyField(title, name, type_str))
    return pretty_fields, external_databases


def str_to_bool(value: str | bool) -> bool:
    if isinstance(value, bool):
        return value
    value = value.lower()
    true_values = ["true", "yes", "y", "1", "on"]
    false_values = ["false", "no", "n", "0", "off"]
    if value in true_values:
        return True
    elif value in false_values:
        return False
    else:
        raise ValueError(f"Cannot convert '{value}' to boolean")


async def create_entity_process_fields(
    fibery_client: FiberyClient, schema: Schema, database: Database, fields: Dict[str, Any]
) -> Tuple[List[Dict[str, Any]], Dict[str, Any]]:
    rich_text_fields = []
    safe_fields = deepcopy(fields)
    for field_name, field_value in fields.items():
        # process rich-text fields
        if database.fields_by_name().get(field_name, None).is_rich_text():
            rich_text_fields.append({"name": field_name, "value": field_value})
            safe_fields.pop(field_name)

        # process workflow fields
        if database.fields_by_name().get(field_name, None).is_workflow():
            if not isinstance(field_value, str):
                raise ValueError(f"Workflow field '{field_name}' should be a string")
        # process enum fields
        field_type = database.fields_by_name().get(field_name, None).type
        if schema.databases_by_name()[field_type].is_enum():
            enum_values_response = await fibery_client.get_enum_values(field_type)
            enum_values = enum_values_response.result
            safe_fields[field_name] = {"fibery/id": next(filter(lambda e: e["Name"] == field_value, enum_values))["Id"]}

    return rich_text_fields, safe_fields



import os
from uuid import uuid4
from typing import List, Dict, Any

import mcp

from fibery_mcp_server.fibery_client import FiberyClient, CommandResponse
from fibery_mcp_server.utils import create_entity_process_fields

create_entities_batch_tool_name = "create_entities_batch"


def create_entities_batch_tool() -> mcp.types.Tool:
    with open(os.path.join(os.path.dirname(os.path.abspath(__file__)), "descriptions", "create_entities_batch"), "r") as file:
        description = file.read()

    return mcp.types.Tool(
        name=create_entities_batch_tool_name,
        description=description,
        inputSchema={
            "type": "object",
            "properties": {
                "database": {
                    "type": "string",
                    "description": "Fibery Database where entities will be created.",
                },
                "entities": {
                    "type": "object",
                    "description": 'List of dictionaries that define what fields to set in format [{"FieldName": value}] (i.e. [{"Product Management/Name": "My new entity"}]).',
                },
            },
            "required": ["database", "entities"],
        },
    )


async def handle_create_entities_batch(fibery_client: FiberyClient, arguments: Dict[str, Any]) -> List[mcp.types.TextContent]:
    database_name: str = arguments.get("database")
    entities: List[Dict[str, Any]] = arguments.get("entities")

    if not database_name:
        return [mcp.types.TextContent(type="text", text="Error: database is not provided.")]

    if not entities or len(entities) == 0:
        return [mcp.types.TextContent(type="text", text="Error: entities is not provided.")]

    schema = await fibery_client.get_schema()
    database = schema.databases_by_name()[database_name]
    if not database:
        return [mcp.types.TextContent(type="text", text=f"Error: database {database_name} was not found.")]

    safe_entities = []
    rich_text_fields_map = {}
    for entity in entities:
        rich_text_fields, safe_entity = await create_entity_process_fields(fibery_client, schema, database, entity)
        safe_entity["fibery/id"] = str(uuid4())
        rich_text_fields_map[safe_entity["fibery/id"]] = rich_text_fields
        safe_entities.append(safe_entity)
    creation_batch_result = await fibery_client.create_entities_batch(database_name, safe_entities)

    if not creation_batch_result.success:
        return [mcp.types.TextContent(type="text", text=str(creation_batch_result))]

    entities_info = []
    for creation_result in creation_batch_result.result:
        creation_result_command = CommandResponse(creation_result["success"], creation_result["result"])
        rich_text_fields = rich_text_fields_map.get(creation_result_command.result["fibery/id"])
        if len(rich_text_fields_map.get(creation_result_command.result["fibery/id"])) > 0:
            secrets_response = await fibery_client.query(
                {
                    "q/from": database_name,
                    "q/select": {
                        field["name"]: [field["name"], "Collaboration~Documents/secret"] for field in rich_text_fields
                    },
                    "q/limit": 1,
                    "q/where": ["=", ["fibery/id"], "$id"],
                },
                {"$id": creation_result_command.result["fibery/id"]},
            )

            for field, secret_response in zip(rich_text_fields, secrets_response.result):
                secret = secret_response.get(field["name"], None)
                if not secret:
                    return [
                        mcp.types.TextContent(
                            type="text", text=f"Error: entity created, but could you populate document {field['name']}"
                        )
                    ]
                doc_result = await fibery_client.create_or_update_document(secret, field["value"])
                if not doc_result.success:
                    return [mcp.types.TextContent(type="text", text=str(doc_result))]

        public_id = creation_result_command.result["fibery/public-id"]
        url = fibery_client.compose_url(database_name.split("/")[0], database_name.split("/")[1], public_id)
        entities_info.append({"id": creation_result_command.result["fibery/id"], "public_id": public_id, "url": url})
    entities_info_str = list(map(lambda ent: f'\nfibery/id: "{ent["id"]}" URL: "{ent["url"]}"', entities_info))
    return [
        mcp.types.TextContent(
            type="text", text=str(f'{len(creation_batch_result.result)} entities created successfully.List of created entities:{entities_info_str}')
        )
    ]



from typing import List

import mcp

from fibery_mcp_server.fibery_client import FiberyClient, Schema, Database

schema_tool_name = "list_databases"


def schema_tool() -> mcp.types.Tool:
    return mcp.types.Tool(
        name=schema_tool_name,
        description="Get list of all databases (their names) in user's Fibery workspace (schema)",
        inputSchema={"type": "object"},
    )


async def handle_schema(fibery_client: FiberyClient) -> List[mcp.types.TextContent]:
    schema: Schema = await fibery_client.get_schema()
    db_list: List[Database] = schema.include_databases_from_schema()

    if not db_list:
        content = "No databases found in this Fibery workspace."
    else:
        content = "Databases in Fibery workspace:\n\n"
        for i, db in enumerate(db_list, 1):
            content += f"{i}. {db.name}\n"

    return [mcp.types.TextContent(type="text", text=content)]



from typing import Dict, Any

import mcp

from ..fibery_client import FiberyClient
from fibery_mcp_server.tools.schema import schema_tool_name, schema_tool, handle_schema
from fibery_mcp_server.tools.database import database_tool_name, database_tool, handle_database
from fibery_mcp_server.tools.query import query_tool_name, query_tool, handle_query
from fibery_mcp_server.tools.current_date import current_date_tool_name, current_date_tool, handle_current_date
from fibery_mcp_server.tools.create_entity import create_entity_tool_name, create_entity_tool, handle_create_entity
from fibery_mcp_server.tools.create_entities_batch import create_entities_batch_tool_name, create_entities_batch_tool, handle_create_entities_batch
from fibery_mcp_server.tools.update_entity import update_entity_tool_name, update_entity_tool, handle_update_entity


def handle_list_tools():
    return [current_date_tool(), schema_tool(), database_tool(), query_tool(), create_entity_tool(), create_entities_batch_tool(), update_entity_tool()]


async def handle_tool_call(fibery_client: FiberyClient, name: str, arguments: Dict[str, Any]):
    if name == schema_tool_name:
        return await handle_schema(fibery_client)
    elif name == database_tool_name:
        return await handle_database(fibery_client, arguments)
    elif name == query_tool_name:
        return await handle_query(fibery_client, arguments)
    elif name == current_date_tool_name:
        return await handle_current_date()
    elif name == create_entity_tool_name:
        return await handle_create_entity(fibery_client, arguments)
    elif name == create_entities_batch_tool_name:
        return await handle_create_entities_batch(fibery_client, arguments)
    elif name == update_entity_tool_name:
        return await handle_update_entity(fibery_client, arguments)
    else:
        return [mcp.types.TextContent(type="text", text=f"Error: Unknown tool {name}")]


__all__ = [
    "handle_list_tools",
    "handle_tool_call",
]



import os
from uuid import uuid4
from typing import List, Dict, Any

import mcp

from fibery_mcp_server.fibery_client import FiberyClient
from fibery_mcp_server.utils import create_entity_process_fields

create_entity_tool_name = "create_entity"


def create_entity_tool() -> mcp.types.Tool:
    with open(os.path.join(os.path.dirname(os.path.abspath(__file__)), "descriptions", "create_entity"), "r") as file:
        description = file.read()

    return mcp.types.Tool(
        name=create_entity_tool_name,
        description=description,
        inputSchema={
            "type": "object",
            "properties": {
                "database": {
                    "type": "string",
                    "description": "Fibery Database where to create an entity.",
                },
                "entity": {
                    "type": "object",
                    "description": 'Dictionary that defines what fields to set in format {"FieldName": value} (i.e. {"Product Management/Name": "My new entity"}).',
                },
            },
            "required": ["database", "entity"],
        },
    )


async def handle_create_entity(fibery_client: FiberyClient, arguments: Dict[str, Any]) -> List[mcp.types.TextContent]:
    database_name: str = arguments.get("database")
    entity: Dict[str, Any] = arguments.get("entity")

    if not database_name:
        return [mcp.types.TextContent(type="text", text="Error: database is not provided.")]

    if not entity:
        return [mcp.types.TextContent(type="text", text="Error: entity is not provided.")]

    schema = await fibery_client.get_schema()
    database = schema.databases_by_name()[database_name]
    if not database:
        return [mcp.types.TextContent(type="text", text=f"Error: database {database_name} was not found.")]
    rich_text_fields, safe_entity = await create_entity_process_fields(fibery_client, schema, database, entity)

    safe_entity["fibery/id"] = str(uuid4())
    creation_result = await fibery_client.create_entity(database_name, safe_entity)

    if not creation_result.success:
        return [mcp.types.TextContent(type="text", text=str(creation_result))]

    if len(rich_text_fields) > 0:
        secrets_response = await fibery_client.query(
            {
                "q/from": database_name,
                "q/select": {
                    field["name"]: [field["name"], "Collaboration~Documents/secret"] for field in rich_text_fields
                },
                "q/limit": 1,
                "q/where": ["=", ["fibery/id"], "$id"],
            },
            {"$id": safe_entity["fibery/id"]},
        )

        for field, secret_response in zip(rich_text_fields, secrets_response.result):
            secret = secret_response.get(field["name"], None)
            if not secret:
                return [
                    mcp.types.TextContent(
                        type="text", text=f"Error: entity created, but could you populate document {field['name']}"
                    )
                ]
            doc_result = await fibery_client.create_or_update_document(secret, field["value"])
            if not doc_result.success:
                return [mcp.types.TextContent(type="text", text=str(doc_result))]

    public_id = creation_result.result["fibery/public-id"]
    url = fibery_client.compose_url(database_name.split("/")[0], database_name.split("/")[1], public_id)
    return [
        mcp.types.TextContent(
            type="text", text=str(f'Entity created successfully. fibery/id: "{safe_entity["fibery/id"]}" URL: "{url}"')
        )
    ]



from typing import List, Dict, Any

import mcp

from fibery_mcp_server.utils import prettify_fields, PrettyField
from fibery_mcp_server.fibery_client import FiberyClient, Schema, Database, Field

database_tool_name = "describe_database"


def database_tool() -> mcp.types.Tool:
    return mcp.types.Tool(
        name=database_tool_name,
        description="Get list of all fields (in format of 'Title [name]: type') in the selected Fibery database and for all related databases.",
        inputSchema={
            "type": "object",
            "properties": {
                "database_name": {
                    "type": "string",
                    "description": "Database name as defined in Fibery schema",
                }
            },
            "required": ["database_name"],
        },
    )


def describe_database(database: str, fields: List[PrettyField]) -> str:
    content = f"Database {database}:\n"
    for field in fields:
        content += f"    {field.title} [{field.name}]: {field.type}\n"
    return content + "\n"


async def handle_database(fibery_client: FiberyClient, arguments: Dict[str, Any]) -> List[mcp.types.TextContent]:
    schema: Schema = await fibery_client.get_schema()

    database_name: str = arguments.get("database_name")
    if not database_name:
        return [mcp.types.TextContent(type="text", text="Error: database_name is not provided.")]

    database: Database | None = schema.databases_by_name().get(database_name, None)
    if not database:
        return [mcp.types.TextContent(type="text", text=f"Error: database {database_name} was not found.")]

    db_fields: List[Field] = database.fields

    if not db_fields:
        return [mcp.types.TextContent(type="text", text="There are no fields found in this Fibery database.")]

    prettified_fields, external_databases = await prettify_fields(
        fibery_client, schema, database, collect_external_databases=True
    )
    external_prettified_databases = [
        (db.name, (await prettify_fields(fibery_client, schema, db))[0]) for db in external_databases
    ]

    content = ""
    for db, fields in [(database.name, prettified_fields)] + external_prettified_databases:
        content += describe_database(db, fields)
    return [mcp.types.TextContent(type="text", text=content)]



from datetime import datetime
from typing import List

import mcp

current_date_tool_name = "current_date"


def current_date_tool() -> mcp.types.Tool:
    return mcp.types.Tool(
        name=current_date_tool_name,
        description="Get today's date in ISO 8601 format (YYYY-mm-dd.HH:MM:SS.000Z)",
        inputSchema={"type": "object"},
    )


async def handle_current_date() -> List[mcp.types.TextContent]:
    date = datetime.now().strftime("%Y-%m-%dT%H:%M:%S.000Z")
    return [mcp.types.TextContent(type="text", text=date)]



import sys
import logging
import asyncio
from typing import List, Dict, Any

import mcp
import click
from mcp.server import Server, NotificationOptions
from mcp.server.models import InitializationOptions

from .fibery_client import FiberyClient
from .tools import handle_list_tools, handle_tool_call

# TODO: add tests for fibery_client and tools
# TODO: add image handlingentity_id for rich-text


async def serve(fibery_host: str, fibery_api_token: str) -> Server:
    server = Server("fibery-mcp-server")
    logging.basicConfig(level=logging.INFO, stream=sys.stderr)
    logger = logging.getLogger("fibery-mcp-server")
    fibery_client = FiberyClient(fibery_host, fibery_api_token)

    @server.list_tools()
    async def list_tools() -> List[mcp.types.Tool]:
        return handle_list_tools()

    @server.call_tool()
    async def call_tool(name: str, arguments: Dict[str, Any]) -> List[mcp.types.TextContent]:
        logger.info(f"Requested tool with uri: {name}")
        try:
            return await handle_tool_call(fibery_client, name, arguments)
        except Exception as e:
            logger.error(f"Tool error: {str(e)}")
            return [mcp.types.TextContent(type="text", text=f"Error: {str(e)}")]

    return server


@click.command()
@click.option(
    "--fibery-host",
    envvar="FIBERY_HOST",
    required=True,
    help="Fibery host (your-account.fibery.io)",
)
@click.option(
    "--fibery-api-token",
    envvar="FIBERY_API_TOKEN",
    required=True,
    help="Fibery API Token",
)
def main(fibery_host: str, fibery_api_token: str) -> None:
    async def _run() -> None:
        async with mcp.stdio_server() as (read_stream, write_stream):
            server = await serve(fibery_host, fibery_api_token)
            await server.run(
                read_stream,
                write_stream,
                InitializationOptions(
                    server_name="Fibery MCP",
                    server_version="0.0.1",
                    capabilities=server.get_capabilities(
                        notification_options=NotificationOptions(),
                        experimental_capabilities={},
                    ),
                ),
            )

    asyncio.run(_run())
